=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed September 2009 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   CurviloftAlong.rb
# Original Date	:   4 Jan 2010 - version 1.0
# Description	:   Specific method for AlongPath methods in the Curviloft script family
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Curviloft

class CVL_LoftAlgo 

#----------------------------------------------------------------------------
# ALONG: analysis of Contours
#----------------------------------------------------------------------------

#Verify contours for Loft Along
def along_check_contours(lst_contours, from_selection=false)

	#Handle only the rail is defined
	nb_contours = lst_contours.length
	return false if nb_contours == 1
	
	#Computing the plates
	lplates = lst_contours.collect { |lpt| plate_create(lpt) }
	lplates.each { |plate| plate_compute_plane plate }
	
	#Testings rails
	along_compute_rail lplates, from_selection

	#Creating the borders for highlight
	@lst_links.each do |link|
		link.hiborders = link.plates.collect { |plate| plate.pts + ((plate.loop) ? [plate.pts[0]] : []) }
	end
		
	true
end

#Calculate the rail information for a given contour irail, vis-a-vis the other contours
def along_compute_rail(lplates, from_selection=false)
	nb_plates = lplates.length - 1
	
	#Computing the intersection of each contours with others
	linfo = []
	range = (from_selection) ? 0..0 : 0..nb_plates
	for irail in range
		pts_rail = lplates[irail].pts
		nr = pts_rail.length - 1
		loop_rail = (pts_rail.first == pts_rail.last)
		
		lcrosses = []
		lplates.each_with_index do |plate, iplate|
			next if iplate == irail
			bary = plate.bary
			llcross = []
			for iseg in 0..nr-1
				lcross = along_intersect_segment_plate pts_rail, iseg, plate
				llcross.push [lcross, lcross[1].distance(bary)] if lcross
			end	
			unless llcross.empty?
				llcross.sort! { |a, b| a[1] <=> b[1] }
				lcrosses.push llcross[0][0]
			end	
		end
		#puts "IRAIL = #{irail} CROSS = #{lcrosses.length}"
		linfo.push [irail, lcrosses]
	end	
	
	#Keeping the contour that has the maximum crosses
	linfo.sort! { |a, b| a[1].length <=> b[1].length } 	
	info = linfo.last
	
	#Selecting the best Rail and creating the ordered list of plates
	irail = info[0]
	@rail = lplates[irail].pts
	#@rail.each { |pt| puts "@@Rail = #{pt}" }
	lcrosses = info[1]
	lcrosses.sort! { |a, b| a[2] <=> b[2] }
	@lst_plates = lcrosses.collect { |a| a[3] }
	#puts "Irail final = #{irail} cross = #{info[1].length}"

	#Only one plate - Generate the plate extremities
	duplicated = false
	noedit = false
	if @lst_plates.length == 1
		plate1 = @lst_plates[0]
		lcross1 = lcrosses[0]
		d, ptinter, iseg = lcross1
		newpt_beg = []
		newpt_end = []
		plate1.pts.each do |pt|
			offpts = G6.curl_offset_by_reference @rail, pt, ptinter, iseg
			newpt_beg.push offpts[0]
			newpt_end.push offpts[-1]
			#puts "\npt = #{pt}"
			#puts "ptbeg = #{offpts[0]}"
			#puts "ptend = #{offpts[-1]}"
		end	
		if plate1.loop
			newpt_beg.push newpt_beg[0]
			newpt_end.push newpt_end[0]
		end

		#puts "iseg = #{iseg} d = #{d} @rail.length = #{@rail.length}"
		#interpts = G6.curl_offset_at_point @rail, ptinter		
		interpts = @rail	
		@lst_plates = []
		#lcrosses = []		
		#unless ptinter == @rail.first
			#puts "ptinetr = #{ptinter}"
			#puts "rail beg = #{@rail[0]}"
			#newpt_beg.each { |pt| puts "New PTBEG = #{pt}" }
			plate_beg = plate_create newpt_beg
			plate_compute_plane plate_beg
			#lcrosses[0, 0] = [0, interpts[0], @rail.length-1, plate_beg]	
			@lst_plates.push plate_beg			
			lcrosses = [[0, interpts[0], 0, plate_beg]] #+ lcrosses	
			#puts "creating BEG Plate lcorsses = #{lcrosses.length}"
		#end

		#@lst_plates.push plate1
		
		#unless ptinter == @rail.last
			plate_end = plate_create newpt_end
			plate_compute_plane plate_end
			lcrosses.push [0, interpts[-1], @rail.length-1, plate_end]
			@lst_plates.push plate_end			
			#puts "creating END Plate lcorsses = #{lcrosses.length}"
		#end
		
		duplicated = true
		noedit = true
		#@lst_plates = [plate_beg, plate_end]
		
	end
	
	#Computing the rail and creating the links
	for i in 0..lcrosses.length-2
		cross1 = lcrosses[i]
		cross2 = lcrosses[i+1]
		plate1 = cross1[3]
		plate2 = cross2[3]
		
		iseg1 = cross1[2]
		iseg2 = cross2[2]
		#puts "cross1[1] = #{cross1[1]}"
		#puts "cross2[1] = #{cross2[1]}"
		rail = (cross1[1] == @rail[iseg1+1]) ? [] : [cross1[1]]
		for i in iseg1+1..iseg2-1
			rail.push @rail[i]
		end
		rail.push cross2[1]
		#puts "RAIL len = #{rail.length}"
		
		plate1.vecrail = rail[0].vector_to rail[1] unless plate1.vecrail
		plate2.vecrail = rail[-2].vector_to rail[-1]
		
		link = link_create plate1, plate2	

		link.rail = rail
		link.rails_native[1] = rail
		link.rails_native[0] = @rail
		link.rails_tot_len = link.rails_native.collect { |rail| G6.curl_length rail }	
		link.duplicated = duplicated
		link.no_edition = noedit
	end	
	@all_duplicated = duplicated

	#puts "final plates = #{@lst_plates.length}"
	
end
 
def along_intersect_segment_plate(pts_rail, iseg, plate)
	#True intersections
	pt1 = pts_rail[iseg]
	ptinter = plate.pts.find { |pt| pt == pt1 }
	return [0, pt1, iseg, plate] if ptinter
	
	pt2 = pts_rail[iseg+1]
	ptinter = plate.pts.find { |pt| pt == pt2 }
	return [0, pt2, iseg+1, plate] if ptinter
	
	#Intersection with the Plate plane
	if plate.plane
		ptinter = G6.intersect_segment_plane pt1, pt2, plate.plane
		#puts "inter in = #{ptinter}"
	else
		line = [plate.pts[0], plate.pts[-1]]
		ptinter = G6.intersect_segment_line pt1, pt2, line
	end
	(ptinter) ? [pt1.distance(ptinter), ptinter, iseg, plate] : nil
end
 
def along_comput_master_directions
	#Parameters for first plate
	plate0 = @lst_plates[0]
	bary0 = plate0.bary
	pts0 = plate0.pts
	ipt0 = 0
	for i in 0..pts0.length-1
		if pts0[i] != bary0
			ipt0 = i
			break
		end	
	end		
	pt0 = pts0[ipt0]
	vec0 = bary0.vector_to pt0
	tr0 = Geom::Transformation.translation vec0
	refpts0 = pts0.collect { |pt| tr0 * pt }
	
	#Reference 2D axes
	
end
 
#----------------------------------------------------------------------------
# ALONG: construction of the shape
#----------------------------------------------------------------------------

#Construct the Along_path curves for a link
def along_junction_construct(link)	
	offsetting = @hprop[:option_offset]
	lbz_pts = []
	link.rail = @rail unless link.rail
	
	#Constructing the junctions by Offset 3D
	if offsetting
		rail = along_oversample_rail link, 0
		normal = @hprop[:offset_normal]
		link.pairs.each do |pair|
			lbz_pts.push G6.curl_offset_by_two_points(rail, pair[0].pt, pair[1].pt, normal)
		end
	
	#Constructing the junction by stretching
	else
		rail = along_oversample_rail link, 1
		link.pairs.each do |pair|
			lbz_pts.push G6.curl_move_extremities(rail, pair[0].pt, pair[1].pt)
		end
	end
	
	lbz_pts
end

#Interpolate the rail (with equal number of points)
#return the rail interpolated and the list of index of points added
def along_oversample_rail(link, irail)
	#No Oversampling
	rail = link.rails_native[irail]
	return rail unless hprop_get(:option_sample, link)
	fsample = hprop_get(:factor_sample, link)
	return rail if fsample < 1

	#Initialization
	dtot = link.rails_tot_len[irail]
	ratio = (dtot > 0) ? 1.0 / dtot : 1.0
	factor = fsample
	
	#Going thorugh each segment of the rails
	newrail = [rail[0]]
	rail_index = []
	len = rail.length-1
	for i in 1..len
		pt_prev = rail[i-1]
		pt = rail[i]
		d = pt.distance(pt_prev)
		
		n = (d * ratio * factor).round
		
		step = 1.0 / (n + 1)
		for j in 1..n
			f = j * step
			newrail.push Geom.linear_combination(1-f, pt_prev, f, pt)
			rail_index.push newrail.length - 1
		end		
		newrail.push pt
	end

	link.rails_sample_index = rail_index
	
	#puts "Computed new rail = #{newrail.length}"
	newrail
end

end	#End Class CVL_LoftAlgo

end	#End Module Curviloft
